/*
 * Copyright 2025 Shihwei Huang <shihwei.huang@focaltech-electronics.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuFocalTouchPlugin, fu_focal_touch_plugin, FU, FOCAL_TOUCH_PLUGIN, FuPlugin)
