/*
 * Copyright 2020 Richard Hughes <richard@hughsie.com>
 * Copyright 2020 H.J. Lu <hjl.tools@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

/* goblint-ignore-next-line: missing_implementation */
void
fu_cpu_helper_cet_testfn1(void);
