/*
 * Copyright 2025 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 * nocheck:expect: file has too many #defined magic values
 */

#define FOO_001 0x123
#define FOO_002 0x123
#define FOO_003 0x123
#define FOO_004 0x123
#define FOO_005 0x123
#define FOO_006 0x123
#define FOO_007 0x123
#define FOO_008 0x123
#define FOO_009 0x123
#define FOO_010 0x123
#define FOO_011 0x123
#define FOO_012 0x123
#define FOO_013 0x123
#define FOO_014 0x123
#define FOO_015 0x123
#define FOO_016 0x123
