/*
 * Copyright 2026 Mario Limonciello <mario.limonciello@amd.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuHpBiosCfgPlugin, fu_hp_bioscfg_plugin, FU, HP_BIOSCFG_PLUGIN, FuPlugin)
