/*
 * Copyright 2026 Sunwinon Electronics Co., Ltd.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuSunwinonHidPlugin, fu_sunwinon_hid_plugin, FU, SUNWINON_HID_PLUGIN, FuPlugin)
