/*
 * Copyright 2024 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include "fu-progress.h"

gdouble
fu_progress_get_global_fraction(FuProgress *self) G_GNUC_NON_NULL(1);
