/*
 * Copyright 2024 Logitech, Inc.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#define FU_LOGITECH_TAP_TOUCH_IC_NAME			 0x2511
#define FU_LOGITECH_TAP_TOUCH_SUPPORTED_PROTOCOL_VERSION 0x03
