/*
 * Copyright 2026 Sean Rhodes <sean@starlabs.systems>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuSunplusCameraPlugin,
		     fu_sunplus_camera_plugin,
		     FU,
		     SUNPLUS_CAMERA_PLUGIN,
		     FuPlugin)
