/*
 * Copyright 2025 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 * nocheck:expect: invalid parameter name self
 */

static gboolean
fu_param_self_firmware(FuFirmware *self)
{
}
