/*
 * Copyright 2026 Himax Company, Ltd.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuHimaxTpPlugin, fu_himax_tp_plugin, FU, HIMAX_TP_PLUGIN, FuPlugin)
