/*
 * Copyright 2024 B&R Industrial Automation GmbH
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuBnrDpPlugin, fu_bnr_dp_plugin, FU, BNR_DP_PLUGIN, FuPlugin)
