/*
 * Copyright 2024 Chris Hofstaedtler <ch@zeha.at>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuRpPicoPlugin, fu_rp_pico_plugin, FU, RP_PICO_PLUGIN, FuPlugin)
