/*
 * Copyright 1999-2023 Logitech, Inc.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuLogitechRallysystemPlugin,
		     fu_logitech_rallysystem_plugin,
		     FU,
		     LOGITECH_RALLYSYSTEM_PLUGIN,
		     FuPlugin)
