// Copyright 2025 Richard Hughes <richard@hughsie.com>
// SPDX-License-Identifier: LGPL-2.1-or-later

// flags to use when configuring debugging
enum FuDumpFlags {
    None = 0,
    ShowAscii = 1 << 0,
    ShowAddresses = 1 << 1,
}
