/*
 * Copyright 2025 Jason Huang <jason.huang@egistec.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuEgisMocPlugin, fu_egis_moc_plugin, FU, EGIS_MOC_PLUGIN, FuPlugin)
