/*
 * Copyright 2025 lazro <2059899519@qq.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuLegionHidPlugin, fu_legion_hid_plugin, FU, LEGION_HID_PLUGIN, FuPlugin)
