// Copyright 2025 Richard Hughes <richard@hughsie.com>
// SPDX-License-Identifier: LGPL-2.1-or-later

#[derive(ToString, FromString)]
enum FuCrcKind {
    Unknown,
    B32Standard,
    B32Bzip2,
    B32Jamcrc,
    B32Mpeg2,
    B32Posix,
    B32Sata,
    B32Xfer,
    B32C,
    B32D,
    B32Q,
    B16Xmodem,
    B16Kermit,
    B16Usb,
    B16Umts,
    B16Tms37157,
    B16Bnr,
    B8Wcdma,
    B8Tech3250,
    B8Standard,
    B8SaeJ1850,
    B8Rohc,
    B8Opensafety,
    B8Nrsc5,
    B8MifareMad,
    B8MaximDow,
    B8Lte,
    B8ICode,
    B8Itu,
    B8Hitag,
    B8GsmB,
    B8GsmA,
    B8DvbS2,
    B8Darc,
    B8Cdma2000,
    B8Bluetooth,
    B8Autosar,
}
