/*
 * Copyright 2025 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 * nocheck:expect: file has too many inline magic values
 */

static void
fu_magic_numbers_inline(void)
{
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
	call(self, 0x123);
}
