/*
 * Copyright 2026 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <glib.h>

/* hash algorithm sizes */
#define FU_TPM_DIGEST_SIZE_SHA1	   20
#define FU_TPM_DIGEST_SIZE_SHA256  32
#define FU_TPM_DIGEST_SIZE_SHA384  48
#define FU_TPM_DIGEST_SIZE_SHA512  64
#define FU_TPM_DIGEST_SIZE_SM3_256 32
