/*
 * Copyright 2026 JS Park <mameforever2@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuLxsTouchPlugin, fu_lxs_touch_plugin, FU, LXS_TOUCH_PLUGIN, FuPlugin)
