/*
 * Copyright 2025 Joe Hong <joe_hung@ilitek.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuIlitekItsPlugin, fu_ilitek_its_plugin, FU, ILITEK_ITS_PLUGIN, FuPlugin)
