/*
 * Copyright 2025 Raydium.inc <Maker.Tsai@rad-ic.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuRaydiumTpPlugin, fu_raydium_tp_plugin, FU, RAYDIUM_TP_PLUGIN, FuPlugin)
