/*
 * Copyright 2023 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuWchCh347Plugin, fu_wch_ch347_plugin, FU, WCH_CH347_PLUGIN, FuPlugin)
