// Copyright 2025 Richard Hughes <richard@hughsie.com>
// SPDX-License-Identifier: LGPL-2.1-or-later

// Flags used to represent report attributes
// Since: 1.9.1
#[derive(ToString(enum), FromString(enum))]
enum FwupdReportFlags {
    // No report flags are set.
    None = 0,
    // Report was generated by the OEM.
    FromOem = 1 << 0,
    // New firmware was newer than the old firmware.
    // Since: 1.9.14
    IsUpgrade = 1 << 1,
    // The report flag is unknown.
    // This is usually caused by a mismatched libfwupdplugin and daemon.
    Unknown = u64::MAX,
}
